#include "window.hpp"

#include "ui_window.h"

#include <QCheckBox>
#include <QGuiApplication>
#include <QStandardItemModel>
#include <fstream>
#include <sys/socket.h>
#include <wayland-client-protocol.h>
#include <wayland-client.h>

static std::string status_to_string(ProtocolStatus status) {
    switch (status) {
        case STABLE:
            return "Stable";
        case STAGING:
            return "Staging";
        case UNSTABLE:
            return "Unstable";
        case WLROOTS:
            return "Wlroots";
        default:
            return "Unknown";
    }
}

// clang-format off
static std::unordered_map<std::string, const Protocol> known_protocols = {
    {"wp_presentation", {STABLE, "Presentation time", "presentation-time"}},
    {"wp_viewporter", {STABLE, "Viewporter", "viewporter"}},
    {"xdg_wm_base", {STABLE, "XDG shell", "xdg-shell"}},

    {"xdg_activation_v1", {STAGING, "XDG activation", "xdg-activation-v1"}},
    {"wp_drm_lease_v1", {STAGING, "DRM lease", "drm-lease-v1"}},
    {"ext_session_lock_manager_v1", {STAGING, "Session lock", "ext-session-lock-v1"}},
    {"wp_single_pixel_buffer_manager_v1", {STAGING, "Single-pixel buffer", "single-pixel-buffer-v1"}},
    {"wp_content_type_manager_v1", {STAGING, "Content type hint", "content-type-v1"}},
    {"ext_idle_notifier_v1", {STAGING, "Idle notify", "ext-idle-notify-v1"}},
    {"wp_tearing_control_manager_v1", {STAGING, "Tearing control", "tearing-control-v1"}},
    {"xwayland_shell_v1", {STAGING, "Xwayland shell", "xwayland-shell-v1"}},
    {"wp_fractional_scale_manager_v1", {STAGING, "Fractional scale", "fractional-scale-v1"}},
    {"wp_cursor_shape_manager_v1", {STAGING, "Cursor shape", "cursor-shape-v1"}},
    {"ext_foreign_toplevel_list_v1", {STAGING, "Foreign toplevel list", "ext-foreign-toplevel-list-v1"}},
    {"wp_security_context_manager_v1", {STAGING, "Security context", "security-context-v1"}},

    {"zwp_fullscreen_shell_v1", {UNSTABLE, "Fullscreen shell", "fullscreen-shell-unstable-v1"}},
    {"zwp_idle_inhibit_manager_v1", {UNSTABLE, "Idle inhibit", "idle-inhibit-unstable-v1"}},
    {"zwp_input_method_context_v1", {UNSTABLE, "Input method", "input-method-unstable-v1"}},
    {"zwp_input_timestamps_manager_v1", {UNSTABLE, "Input timestamps", "input-timestamps-unstable-v1"}},
    {"zwp_keyboard_shortcuts_inhibit_manager_v1", {UNSTABLE, "Keyboard shortcuts inhibit", "keyboard-shortcuts-inhibit-unstable-v1"}},
    {"zwp_linux_dmabuf_v1", {UNSTABLE, "Linux DMA-BUF", "linux-dmabuf-unstable-v1"}},
    {"zwp_linux_explicit_synchronization_v1", {UNSTABLE, "Linux explicit synchronization", "linux-explicit-synchronization-unstable-v1"}},
    {"zwp_pointer_constraints_v1", {UNSTABLE, "Pointer constraints", "pointer-constraints-unstable-v1"}},
    {"zwp_pointer_gestures_v1", {UNSTABLE, "Pointer gestures", "pointer-gestures-unstable-v1"}},
    {"zwp_primary_selection_device_manager_v1", {UNSTABLE, "Primary selection", "primary-selection-unstable-v1"}},
    {"zwp_relative_pointer_manager_v1", {UNSTABLE, "Relative pointer", "relative-pointer-unstable-v1"}},
    {"zwp_tablet_manager_v2", {UNSTABLE, "Tablet", "tablet-unstable-v2"}},
    {"zwp_text_input_v3", {UNSTABLE, "Text input", "text-input-unstable-v3"}},
    {"zxdg_decoration_manager_v1", {UNSTABLE, "XDG decoration", "xdg-decoration-unstable-v1"}},
    {"zxdg_exporter_v2", {UNSTABLE, "XDG foreign", "xdg-foreign-unstable-v2"}},
    {"zxdg_output_manager_v1", {UNSTABLE, "XDG output", "xdg-output-unstable-v1"}},
    {"zwp_xwayland_keyboard_grab_manager_v1", {UNSTABLE, "Xwayland keyboard grabbing", "xwayland-keyboard-grab-unstable-v1"}},

    {"zwlr_data_control_manager_v1", {WLROOTS, "Data control", "wlr-data-control-unstable-v1"}},
    {"zwlr_export_dmabuf_manager_v1", {WLROOTS, "Export DMA-BUF", "wlr-export-dmabuf-unstable-v1"}},
    {"zwlr_foreign_toplevel_manager_v1", {WLROOTS, "Foreign toplevel management", "wlr-foreign-toplevel-management-unstable-v1"}},
    {"zwlr_gamma_control_manager_v1", {WLROOTS, "Gamma control", "wlr-gamma-control-unstable-v1"}},
    {"zwlr_input_inhibit_manager_v1", {WLROOTS, "Input inhibitor", "wlr-input-inhibitor-unstable-v1"}},
    {"zwlr_layer_shell_v1", {WLROOTS, "Layer shell", "wlr-layer-shell-unstable-v1"}},
    {"zwlr_output_manager_v1", {WLROOTS, "Output management", "wlr-output-management-unstable-v1"}},
    {"zwlr_output_power_manager_v1", {WLROOTS, "Output power management", "wlr-output-power-management-unstable-v1"}},
    {"zwlr_screencopy_manager_v1", {WLROOTS, "Screencopy", "wlr-screencopy-unstable-v1"}},
    {"zwlr_virtual_pointer_v1", {WLROOTS, "Virtual pointer", "wlr-virtual-pointer-unstable-v1"}},
};
// clang-format on

static pid_t pid_from_fd(int fd) {
    struct ucred ucred;
    socklen_t len = sizeof(struct ucred);
    if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) == -1) {
        perror("getsockopt failed");
        exit(1);
    }
    return ucred.pid;
}

static std::string process_name_from_pid(const pid_t pid) {
    std::string procpath = QString::asprintf("/proc/%d/comm", pid).toStdString();

    std::ifstream infile(procpath);
    if (infile.is_open()) {
        std::string out;
        std::getline(infile, out);
        infile.close();
        return out;
    } else {
        return "Unknown";
    }
}

static void registry_global(void* data, wl_registry* registry, uint32_t name, const char* interface, uint32_t version) {
    (void) registry;
    (void) name;

    if (std::string(interface).starts_with("wl_")) {
        return;
    }

    auto* window = static_cast<Window*>(data);
    window->newGlobal(interface, version);
}

static void registry_global_remove(void* data, struct wl_registry* registry, uint32_t name) {
    (void) data;
    (void) registry;
    (void) name;
}

Window::Window(QNativeInterface::QWaylandApplication* waylandApp, QWidget* parent)
    : QMainWindow(parent), model(QStandardItemModel(0, 4)), ui(new Ui::Window) {
    ui->setupUi(this);

    model.setHorizontalHeaderLabels({"Status", "Name", "Identifier", "Implementation"});

    ui->tableView->setModel(&model);
    ui->tableView->setSortingEnabled(true);
    ui->tableView->horizontalHeader()->setSectionResizeMode(1, QHeaderView::ResizeToContents);
    ui->tableView->horizontalHeader()->setSectionResizeMode(2, QHeaderView::Stretch);
    ui->tableView->horizontalHeader()->setSectionResizeMode(3, QHeaderView::ResizeToContents);
    ui->tableView->horizontalHeader()->setSectionResizeMode(4, QHeaderView::ResizeToContents);

    for (auto protocol : known_protocols) {
        addProtocol(protocol.second);
    }

    ui->tableView->sortByColumn(0, Qt::AscendingOrder);

    auto* display = waylandApp->display();
    auto fd = wl_display_get_fd(display);
    auto pid = pid_from_fd(fd);
    auto pname = process_name_from_pid(pid);
    ui->compositor->setText(QString::asprintf("Compositor: %s", pname.c_str()));

    auto listener = wl_registry_listener{.global = registry_global, .global_remove = registry_global_remove};

    auto* registry = wl_display_get_registry(display);
    wl_registry_add_listener(registry, &listener, (void*) this);
    wl_display_roundtrip(display);
}

Window::~Window() {
    delete ui;
}

void Window::newGlobal(const std::string interface, const uint32_t version) {
    const Protocol& protocol = known_protocols[interface];

    if (protocol.status != UNKNOWN) {
        auto matches = model.findItems(QString::fromStdString(protocol.name), Qt::MatchExactly, 2);

        for (auto match : matches) {
            auto versionItem = model.item(match->row(), 3);
            versionItem->setCheckState(Qt::Checked);
            versionItem->setText(QString::asprintf("%d", version));

            for (auto i = 0; i < model.columnCount(); i++) {
                auto item = model.item(match->row(), i);
                auto bg = item->background();
                bg.setStyle(Qt::NoBrush);
                item->setBackground(bg);
            }
        }
    }
}

void Window::addProtocol(const Protocol& protocol) {
    auto statusItem = new QStandardItem(QString::fromStdString(status_to_string(protocol.status)));
    statusItem->setTextAlignment(Qt::AlignCenter);

    auto versionItem = new QStandardItem(false);
    versionItem->setCheckState(Qt::Unchecked);
    versionItem->setText("N/A");

    auto items = QList<QStandardItem*>({
        statusItem,
        new QStandardItem(QString::fromStdString(protocol.pretty)),
        new QStandardItem(QString::fromStdString(protocol.name)),
        versionItem,
    });

    for (auto item : items) {
        auto bg = item->background();
        bg.setStyle(Qt::Dense4Pattern);
        item->setBackground(bg);
    }

    model.appendRow(items);
}
